#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>

//#define DEBUG 1

typedef unsigned long long ull;
struct ver
{
    char deg;
    ull id;
    char *next;
    char par;
};

typedef struct ver ver;

struct tree
{
    char size;
    ver* list;
};

typedef struct tree tree;

struct tList
{
    tree *t;
    char n;
    struct tList *next;
};

typedef struct tList tList;

struct pair
{
    char a;
    char b;
};

typedef struct pair pair;

ver *CMP;
/* global link for debug reasons */
FILE* dbg1;
/* counter for non-isomorphic trees */
ull trees;

/* compare functions for pairs */
#define A(x) (*(const pair *) x).a
#define B(x) (*(const pair *) x).b
#define MIN(x,y) ((x < y) ? x : y)
int smallercomp(const void *x, const void *y)
{
    return MIN(A(x),B(x)) - MIN(A(y),B(y));
}

/* debug functions for printing subtrees, and tree lists */
void print_vertex(FILE *f,tree *t, char vertex)
{
    fprintf(f, "vertex %d (id:%llu) -> %d <- [", vertex, t->list[vertex].id, t->list[vertex].par);
    for(int i=0; i<t->list[vertex].deg;i++)
    {
        fprintf(f, "%d, ", t->list[vertex].next[i]);
    }
    fprintf(f, "]\n");
    for(int i=0; i<t->list[vertex].deg;i++)
    {
        print_vertex(f, t, t->list[vertex].next[i]);
    }
}

void print_tlist(FILE *f, tList *t)
{
    fprintf(f, "printing subtree rooted at %d\n", t->n);
    print_vertex(f, t->t, t->n);
}

void print_list(FILE *f, tList *t)
{
    if(t == NULL)
        return;
    
    print_tlist(f, t);
    //fprintf(f, "next is %d\n", t->next);
    print_list(f, t->next);

}

void erase_tree(tree *t)
{
    for(int i=0;i<t->size;i++)
    {
        free(t->list[i].next);
    }
    free(t->list);
    free(t);
}

// erases a tList, not the whole list
void erase_tList(tList *a)
{
    if(a == NULL) return;
    //erase_tree(a->t);
    free(a);
}
// checks if two subtrees are different
// the subtrees have to have sorted arrays of children

// returns 'different' if any of those is NULL
char different(tList *a, tList *b)
{
    if(a == NULL || b == NULL)
        return 1;
    
    /* fprintf(stderr, "comparing nodes %d and %d", a->n,b->n);
    print_tlist(stderr, a);
    print_tlist(stderr, b);*/
    
    if(a->t->list[a->n].deg != b->t->list[b->n].deg)
    {
#ifdef DEBUG
        fprintf(stderr, "different: degrees\n");
#endif
        return 1;
    }
    for(char i=0; i<a->t->list[a->n].deg;i++)
    {
        char an = a->t->list[a->n].next[i];
        char bn = b->t->list[b->n].next[i];
        if(a->t->list[an].id != b->t->list[bn].id)
        {
#ifdef DEBUG
            fprintf(stderr, "different: id of a:%llu is not id of b:%llu\n", a->t->list[an].id, b->t->list[bn].id);
#endif
            return 1;
        }
    }
    return 0;
}

/* compares identifiers for two vertices
 * since vertices are chars, we need a tree to compare
 * it with */
int compareIds(const void *a, const void *b)
{
    ull A = CMP[(*(char *)a)].id;
    ull B = CMP[(*(char *)b)].id;
    if(A>B)
        return 1;
    if(A<B)
        return -1;
    return 0;
}


// adds a in front of a list
// warning: a->next is lost forever
void add(tList *a, tList **list)
{
    a->next = (*list);
    (*list) = a;
    return;
}

void aruAdd(ull **array, ull element, ull *max, ull *cap)
{
    if(*max == (*cap))
    {
        if(*cap == 0)
            (*cap) = 2;
        else
            (*cap) *= 2;

        *array = realloc(*array, (*cap)*sizeof(ull));
    }

    (*array)[(*max)++] = element;
}

void arcAdd(char **array, char element, char *max, char *cap)
{
#ifdef DEBUG
    fprintf(stderr, "adding %d to an array[%d], capacity %d\n", (int) element, (int) *max, (int) *cap);
#endif
    if(*max == (*cap))
    {
        if(*cap == 0)
            (*cap) = 2;
        else
            (*cap) *= 2;
        *array = realloc(*array, (*cap)*sizeof(char));
    }

    (*array)[(*max)++] = element;
}

char dpass(tList *treeList, char n, tList **depths, char maxdeg)
{
#ifdef DEBUG
    fprintf(stderr, "one pass of depthsort, vertex number %d\n",n);
    print_tlist(dbg1, treeList);
#endif
    tList *ret = malloc(sizeof(tList));
    ret->t = treeList->t;
    ret->n = n;
    char depth = 0, dcur;
    if(ret->t->list[n].deg == 0)
    {
        ret->t->list[n].id = 1;
        free(ret);
        return 1;
    }

    for(char i=0; i<ret->t->list[n].deg; i++)
    {
        dcur = dpass(treeList, ret->t->list[n].next[i], depths, maxdeg);
#ifdef DEBUG
        fprintf(stderr, "back to vertex %d, depth %d \n", n, dcur);
#endif
        depth = ((dcur > depth) ? dcur : depth);

    }
#ifdef DEBUG
    fprintf(stderr, "vertex %d final depth %d\n", n, depth);
#endif
    add(ret, &(depths[++depth]));
    return depth;
}



void depthBucket(tList *treeList, tList **depths, char maxdeg)
{
    tList *cur = treeList, *next;
    while(cur != NULL)
    {
        next = cur->next;
        dpass(cur, cur->n, depths, maxdeg);
        cur = next;
    }
}



void oneBucket(tList *treeList, char maxdeg, tList **bucket, ull *buccap, ull *bucmax, ull *total)
{

    // first pass -- sort every vertex's children
    // according to id
    tList *cur = treeList;
    while(cur != NULL)
    {
        CMP=cur->t->list;
        qsort(cur->t->list[cur->n].next, cur->t->list[cur->n].deg, sizeof(char), compareIds);

#ifdef DEBUG
        for(char i=0; i<cur->t->list[cur->n].deg;i++)
        {
            char k = cur->t->list[cur->n].next[i];
            fprintf(stderr," %llu", cur->t->list[k].id);
        }
        fprintf(stderr, "\n");
#endif
        cur = cur->next;
     }

    tList *vals, *next, *last;
    ull nonmax = 0;
    ull noncap = 1000;

    ull *nonempty = malloc(noncap*sizeof(ull));
    // now the bucket sort itself
    char pass = 0;
    vals = treeList;
    ull curid;

    char iteration = 1;
    // put into buckets
    while(vals != NULL)
    {
#ifdef DEBUG
        fprintf(stderr, "dumping into buckets\n");
        fprintf(stderr, "vals are:\n");
        print_list(stderr, vals);
        fprintf(stderr, "----\n");
#endif
        cur = vals;
        
        while(cur != NULL)
        {
            curid = cur->t->list[cur->t->list[cur->n].next[iteration-1]].id;
            assert(curid != 0);
            if(curid >= *buccap)
            {
                ull newlim = (curid > 2*(*buccap) ? curid : 2*(*buccap));
                bucket = realloc(bucket, newlim*sizeof(tList *));
                *buccap = newlim;
                for(ull c = *buccap; c<newlim; c++)
                {
                    bucket[c] = NULL;
                }
            }

            next = cur->next;
            if(bucket[curid] == NULL)
            {
#ifdef DEBUG
                fprintf(stderr, "bucket[%llu] is now non-empty\n", curid);
#endif
                aruAdd(&nonempty, curid, &nonmax, &noncap);
            }
#ifdef DEBUG
            fprintf(stderr, "adding cur to bucket[%llu]\n", curid);
#endif
            add(cur, &(bucket[curid]));
            cur = next;
        }
        vals = NULL;
        // collect from buckets
#ifdef DEBUG
        fprintf(stderr, "gathering from buckets, last nonempty is %llu\n", nonmax);
#endif
        for(ull i=0; i<nonmax;i++)
        {
            cur = bucket[nonempty[i]];
#ifdef DEBUG
            fprintf(stderr, "gathering from %llu\n", nonempty[i]);
#endif
            last = NULL;
            while(cur != NULL)
            {
                if(iteration == cur->t->list[cur->n].deg)
                {   
                    //fprintf(stderr, "is last NULL? %d\n", last==NULL);
                    if(!different(last,cur))
                    {
#ifdef DEBUG
                        fprintf(stderr, "the same, labeling tree:\n");
                        print_tlist(stderr, cur);
#endif
                        cur->t->list[cur->n].id = last->t->list[last->n].id;
#ifdef DEBUG
                        fprintf(stderr, "with id %llu\n", last->t->list[last->n].id);
#endif
                    } else {
                        cur->t->list[cur->n].id = ++(*bucmax);
#ifdef DEBUG
                        fprintf(stderr, "different, new id %llu\n", cur->t->list[cur->n].id);
#endif
                        // if it's a root, add a global counter of different trees
                        if(!cur->n)
                        {
#ifdef DEBUG
                            fprintf(stderr, "trees++;\n");
#endif
                            trees++;
                        }
                    }

                    // last is not needed, erase it and update it with cur
                    next = cur->next;
                    erase_tList(last);
                    last = cur;
                }
                else
                {
#ifdef DEBUG
                    fprintf(stderr, "larger degree, needs another pass\n");
#endif
                    next = cur->next;
                    //fprintf(stderr, "cur is %d, vals is %d\n", cur, vals);
                    add(cur,&vals);
#ifdef DEBUG
                    print_list(stderr, vals);
#endif
                }
                cur = next;
            }
            erase_tList(last);
            bucket[nonempty[i]] = NULL;
        }
        nonmax=0;
        iteration++;
        //fprintf(stderr, "new iteration\n");
        //print_list(stderr, vals);
        //fprintf(stderr, "-------------------\n");
        //fprintf(stderr, "%d\n", vals == NULL);
    }
#ifdef DEBUG
    fprintf(stderr, "Ending one bucket pass.\n");
#endif
    free(nonempty);
}

// sorts trees using a sophisticated bucket sort,
// assumes all trees have the same cardinality
// returns a ull -- number of non-isomorphic trees
ull countTrees(tList *treeList, char maxdeg)
{
    tList *depths[maxdeg+1];
    for(char i=0;i<=maxdeg;i++)
    {
        depths[i] = NULL;
    }
    ull total = 0;
    ull buccap = 1000;
    ull bucmax = 1;
#ifdef DEBUG
    fprintf(stderr, "starting depthsort\n");
#endif

    depthBucket(treeList, depths, maxdeg);

#ifdef DEBUG
    fprintf(stderr, "end depthsort, result:\n");
#endif    

    tList **bucket = calloc(buccap, sizeof(tList *));
    for(char i=2;i<=maxdeg;i++)
    {
        tList *d = depths[i];
#ifdef DEBUG
        fprintf(stderr, "depth %d\n", i);
        print_list(stderr, d);
        fprintf(stderr, "---\n");
#endif
        oneBucket(d, maxdeg, bucket, &buccap, &bucmax, &total);
    }

    return total;
}

// loads rooted trees from the McKay format
// 0 -- the root
tList *loadTrees(FILE *fin, int vertices)
{
#ifdef DEBUG
    fprintf(stderr, "loading trees on %d vertices\n", vertices);
#endif
    tList *ret = NULL, *cur = NULL;
    char **nei, *neicur, *neicap;
    char *queue;
    nei = malloc(vertices*sizeof(char *));
    neicur = calloc(vertices, sizeof(char));
    neicap = calloc(vertices, sizeof(char));
    queue = malloc(vertices*sizeof(char));
    int a,b,c,m,j;
    fscanf(fin, " ");
    while(!feof(fin))
    {
        cur = malloc(sizeof(tList));
        cur->t = malloc(sizeof(tree));
        cur->t->list = malloc(vertices*sizeof(ver));
        assert(cur->t->list != NULL);
        cur->n = 0;
        cur->t->size = vertices;
        cur->next = NULL; 
        for(char i=0;i<vertices;i++)
        {
            cur->t->list[i].next = NULL;
            cur->t->list[i].deg = 0;
            cur->t->list[i].par = -1;
            neicur[i] = neicap[i] = 0;
        }
        
        for(char i=0;i<vertices-1;i++)
        {
            fscanf(fin, " %d %d", &a, &b); a--; b--;
            arcAdd(&nei[b],a,&neicur[b],&neicap[b]);
            arcAdd(&nei[a],b,&neicur[a],&neicap[a]);
        }
        fscanf(fin, " \n");
        queue[0] = 0;
        m=1;
        assert(cur->t->list != NULL);
        for(char i=0;i<vertices;i++)
        {
            j = queue[i];
            if( j == 0)
            {
                cur->t->list[j].next = malloc(neicur[j] * sizeof(char));
                cur->t->list[j].deg = neicur[j];
            }
            else
            {
                cur->t->list[j].next = malloc((neicur[j]-1) * sizeof(char));
                cur->t->list[j].deg = neicur[j]-1;
            }
#ifdef DEBUG
            fprintf(stderr, "creating neighbourhood of %d\n", j);
#endif
            c = 0;
            for(int k=0;k<neicur[j];k++)
            {
                if(cur->t->list[nei[j][k]].par != -1 || nei[j][k] == 0)
                {
#ifdef DEBUG
                    fprintf(stderr, "root of %d is %d\n", j, nei[j][k]);
#endif
                    cur->t->list[j].par = nei[j][k];
                } else {
#ifdef DEBUG
                    fprintf(stderr, "adding %d as son of %d\n", nei[j][k], j);
#endif
                    cur->t->list[j].next[c++] = nei[j][k];
                    cur->t->list[nei[j][k]].par = j;
                    queue[m++] = nei[j][k];
                }
            }
        }
        for(int i=0; i<vertices;i++)
        {
            free(nei[i]);
            nei[i] = NULL;
        }
        //print_tlist(cur);
        add(cur, &ret);
    }
    free(neicur);
    free(neicap);
    free(queue);
    free(nei);
    return ret;
}

int main(int argc, char **argv)
{
    tList *a = NULL;
    FILE *fin;
    FILE *dbg2;
    trees = 0;
    fin = fopen(argv[1], "r");
    if(fin == NULL)
    {
        fprintf(stderr, "Target file cannot be opened.\n");
        return -1;
    }
    a = loadTrees(fin, atoi(argv[2]));
#ifdef DEBUG
    dbg2 = fopen("load.log", "w");
    print_list(dbg2,a);
    dbg1 = fopen("dpass.log", "w");
#endif
    if(a != NULL)
        countTrees(a, atoi(argv[2]));
    fprintf(stdout, "%llu non-isomorphic trees found.\n", trees);
#ifdef DEBUG
    fclose(dbg1);
#endif
    return 0;
}
