#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

//#define DEBUG 1

typedef unsigned long long int ull;
typedef ull ednum;
typedef ull venum;

struct edge{
    venum f;
    venum t;
    ednum nextf;
    ednum nextt;
    int sel;
};

typedef struct edge edge;

struct graph{
    edge *elist;
    ednum *fr;
    ednum *to;
    ull v;
    ull e;
    int limit;
    int *vsel;

    venum *leaves;
    int curlf; 
};

typedef struct graph graph;

int Treesize; // the desired size of the trees
// output file pointers
FILE* Normalsized;
FILE* Subdivided; 
void eraseg(graph *g) 
{
    free(g->elist);
    free(g->fr);
    free(g->to);

    if(g->vsel != NULL)
        free(g->vsel);
    if(g->leaves != NULL)
        free(g->leaves);
    free(g);
}

// load from mckay's "readable" format
// assumes file is already open
graph* loadmckay(FILE* fin)
{
    venum a=0,b=0;
    graph* r = malloc(sizeof(graph));
    fscanf(fin, " %llu %llu", &a, &b);
    r->v = a;
    r->e = b;
    r->leaves = NULL;
    r->vsel = NULL;
    r->fr = malloc((r->v+1) * sizeof(ednum));
    r->to = malloc((r->v+1) * sizeof(ednum));
    r->elist = malloc((r->e+1) * sizeof(edge));
    //memset(r->fr, 0, (r->v+1) * sizeof(ednum));
    //memset(r->to, 0, (r->v+1) * sizeof(ednum));
    for(ednum j=0;j<r->v+1;j++)
    {
        r->fr[j] = 0;
        r->to[j] = 0;
    }

    //for(ednum j=1;j<=r->v;++)
    //    printf("r->to[%llu] = %llu", j, r->to[j]);

    for(ednum j=1;j<=r->e; j++)
    {
        fscanf(fin, " %llu %llu", &a, &b);
#ifdef DEBUG
        fprintf(stderr, "loaded edge %llu %llu\n", a,b);
#endif
        a++; b++;
        r->elist[j].f = a;
        r->elist[j].t = b;
        r->elist[j].nextf = r->fr[a];
        r->fr[a] = j;
        r->elist[j].nextt = r->to[b];
        r->to[b] = j;
    }
    fscanf(fin, " ");
    return r;
}

void debug(graph *g)
{
    fprintf(stderr, "d: |G| = %llu\n", g->v);
    fprintf(stderr, "d: ||G|| = %llu\n", g->e);
    fprintf(stderr, "d: vertex check\n");
    for(venum j=1;j<=g->v; j++)
    {
        fprintf(stderr, "d: %llu firstfrom %llu firstto %llu\n",j,g->fr[j],
                g->to[j]);
    }
    fprintf(stderr, "d: edge check\n");
    for(ednum j=1;j<=g->e; j++)
    {
        fprintf(stderr, "d: %llu: %llu -> %llu (nf: %llu nt: %llu)\n", j,  g->elist[j].f, g->elist[j].t,
                g->elist[j].nextf, g->elist[j].nextt);
    }

    fprintf(stderr, "d: leaf check:\n");
    
    if(g->leaves == NULL) return;

    for(venum i=0;i<g->curlf;i++)
    {
        fprintf(stderr, " %llu", g->leaves[i]);
    }
    fprintf(stderr, "\n");
}

void output_tree(graph *g)
{
#ifdef DEBUG
    fprintf(stderr, "starting output\n");
#endif
    // output larger (subdivided) trees elsewhere
    FILE *out = Normalsized;
    if((g->e+1) > Treesize)
        out = Subdivided;

    for(int i=1;i<=g->e;i++)
    {
        fprintf(out, "%llu %llu  ", g->elist[i].f, g->elist[i].t);
    }
    fprintf(out, "\n");
}

graph *CG;

void print_tuple(int *array, int k)
{
    int i=0;
    for(i=0;i<k;i++)
        fprintf(stderr, " %d ", array[i]);
    fprintf(stderr, "\n");
}
/* e is the edge we just came through */
int dfs_check(graph *g, venum vertex, ednum le)
{
#ifdef DEBUG
    fprintf(stderr, "recursing on vertex %llu\n", vertex);
#endif
    g->vsel[vertex] = 1;
    int a=1, b=1;
    int checked = 0;
    venum ov;
    for(ednum e = g->fr[vertex]; e!=0; e = g->elist[e].nextf)
    {
        checked++;
#ifdef DEBUG
        fprintf(stderr, "checking edge %llu from %llu to %llu\n", e, g->elist[e].f, g->elist[e].t);
#endif
        if(e == le)
            continue;

        ov = g->elist[e].t;
        if(g->vsel[ov] == 1)
        {
#ifdef DEBUG
            fprintf(stderr, "DFS found vertex %llu has a back edge with %llu\n", vertex, ov);
#endif
            return 0;
        }
        else
        {
            a = dfs_check(g, ov, e);
            if(!a) return 0;
        }
    }
    
    for(ednum e = g->to[vertex]; e!=0; e = g->elist[e].nextt)
    {
        checked++;
        //fprintf(stderr, "checking edge %llu from %llu to %llu\n", e, g->elist[e].f, g->elist[e].t);
        //fprintf(stderr, "next edge is %llu\n",  g->elist[e].nextt);
        if(e == le)
            continue;

        ov = g->elist[e].f;
        if(g->vsel[ov] == 1)
        {
#ifdef DEBUG
            fprintf(stderr, "DFS found vertex %llu has a back edge with %llu\n", vertex, ov);
#endif
            return 0;
        }
        else
        {
            b = dfs_check(g, ov, e);
            if(!b) return 0;
        }
    }

    /* we only care for leaves if it's a tree, so it's okay to report here*/
    if(checked == 1)
        g->leaves[g->curlf++] = vertex;

    return 1;
}

/* a simple check if the graph is connected after DFS is run
 * on it */

int conn_check(graph *g)
{
#ifdef DEBUG
    fprintf(stderr, "Starting connected check.\n");
#endif
    for(int v=1; v<= g->v; v++)
    {
        if(!g->vsel[v])
            return 0;
    }

    return 1;
}


/* validate the graph -- in our case, make
 * sure it's a tree */
int valid(graph *g)
{
#ifdef DEBUG
    fprintf(stderr, "Starting inducedness check.\n");
#endif
    int ret = dfs_check(g, 1, 0);
    if(!ret)
        return 0;
    else
        return conn_check(g);
}


void init_graph(graph *ng, venum v)
{
    ng->v = v;
    ng->e = 0;
    /* TODO elist allocation (and later reallocation) */
    ng->elist = malloc(3*sizeof(edge));
    ng->fr = calloc((ng->v+1), sizeof(ednum));
    ng->to = calloc((ng->v+1), sizeof(ednum));
    ng->vsel = calloc((ng->v+1),sizeof(int));
    
    /* for determining leaves */
    ng->leaves = calloc((ng->v+1), sizeof(venum));
    ng->curlf = 0;
    ng->limit = 3;
}

void add_edge(graph *ng, venum fr, venum to)
{
#ifdef DEBUG
    fprintf(stderr, "creating edge from %llu to %llu\n", fr, to);
#endif

    ng->e++;
    if(ng->e == ng->limit)
    {
        ng->limit *= 2;
        ng->elist = realloc(ng->elist, ng->limit * sizeof(edge));
        assert(ng->elist != NULL);
    }

    ng->elist[ng->e].f = fr;
    ng->elist[ng->e].t = to;
    ng->elist[ng->e].nextf = ng->fr[fr];
    ng->elist[ng->e].nextt = ng->to[to];
    ng->fr[fr] = ng->e;
    ng->to[to] = ng->e;
}

/* replaces edge (fr,to) with an edge either (fr,newdest) or
 * (newdest, to), returns the other vertex */

venum replace_edge(graph *g, venum fr, venum to, venum newdest)
{
#ifdef DEBUG
    fprintf(stderr, "d: rerouting edge (%llu,%llu) to %llu\n", fr, to, newdest);
#endif
    ednum i=0, j=0, e, prev;
    int found = 0;
    for(i = g->fr[fr]; i != 0; i = g->elist[i].nextf)
    {
        if(g->elist[i].t == to)
        {
            e = i;
            prev = 0;
            found = 1;
            for(j = g->to[to]; j != 0; j = g->elist[j].nextt)
            {
                if(j == e)
                {
                    if(!prev)
                    {
                        g->to[to] = g->elist[e].nextt;
                    } else {
                        g->elist[prev].nextt = g->elist[e].nextt;
                    }
                    prev = j;
                }
            }
            g->elist[i].t = newdest;
            g->elist[i].nextt = g->to[newdest];
            g->to[newdest] = i;
        }

        if(found)
        {
#ifdef DEBUG
            fprintf(stderr, "returning %llu\n", to);
#endif
            return to;
        }
    }

    for(i = g->fr[to]; i != 0; i = g->elist[i].nextf)
    {
        //printf("i = %llu\n", i);
        if(g->elist[i].t == fr)
        {
            e = i;
            prev = 0;
            found = 1;
            for(j = g->to[fr]; j != 0; j = g->elist[j].nextt)
            {
                //printf("j = %llu\n", j);
                if(j == e)
                {
                    if(!prev)
                    {
                        g->to[fr] = g->elist[e].nextt;
                    } else {
                        g->elist[prev].nextt = g->elist[e].nextt;
                    }
                    prev = j;
                }
            }
            g->elist[i].t = newdest;
            g->elist[i].nextt = g->to[newdest];
            g->to[newdest] = i;
        }

        if(found)
        {
#ifdef DEBUG
            fprintf(stderr, "returning %llu\n", fr);
#endif
            return fr;
        }
    }
}

/* just relabel vertices a and b -- effectively switch the
 * two numbers in the whole graph
 */
void swap_vertices(graph *g, venum a, venum b)
{
    ednum i;
    for(i = 1; i <= g->e; i++)
    {
        if(g->elist[i].f == a)
            g->elist[i].f = b;
        else if(g->elist[i].f == b)
            g->elist[i].f = a;

        if(g->elist[i].t == a)
            g->elist[i].t = b;
        else if(g->elist[i].t == b)
            g->elist[i].t = a;
    }

    i = g->fr[a];
    g->fr[a] = g->fr[b];
    g->fr[b] = i;

    i = g->to[a];
    g->to[a] = g->to[b];
    g->to[b] = i;

}

int degree(graph *g, venum v)
{
    int r =0;
    for(ednum i = g->fr[v]; i !=0; i = g->elist[i].nextf)
        r++;
    for(ednum i = g->to[v]; i !=0; i = g->elist[i].nextt)
        r++;
    return r;
}

graph *G;

/* finds the center and switches its number
 * with '1'
 */

void centralize(graph *g)
{
    venum v, swap;
    for(int i=1; i<=g->v; i++)
    {
        g->vsel[i] = 0;
    }
    int *height, curq=0;
    venum *queue;
    height = calloc(g->v, sizeof(int));
    queue = calloc(g->v, sizeof(venum));
    for(int i=1; i<=g->v;i++)
    {
        // calculate the degree of every vertex
        g->vsel[i] = degree(g, i);

        if(degree(g, i) == 1)
        {
            height[curq] = 1;
            queue[curq++] = i;
#ifdef DEBUG
            fprintf(stderr, "d: setting vertex %d as leaf\n", i);
#endif
            g->vsel[i] = -1;
        } 
    }
    
    for(int i=0;i<g->v;i++)
    {
#ifdef DEBUG
        fprintf(stderr, "center checking vertex %llu of height %d\n", queue[i], height[i]);
#endif
        v = queue[i];
        for(ednum e = g->fr[v]; e !=0; e = g->elist[e].nextf)
        {
            if(g->vsel[g->elist[e].t] > 0)
                g->vsel[g->elist[e].t]--;
            if(g->vsel[g->elist[e].t] > -1 && g->vsel[g->elist[e].t] <= 1)
            {
                g->vsel[g->elist[e].t] = -1;
                height[curq] = height[i]+1;
#ifdef DEBUG
                fprintf(stderr, "the %d vertex in the queue is high %d\n", curq, height[v]+1);
#endif
                queue[curq++] = g->elist[e].t;
            }
        }

        for(ednum e = g->to[v]; e !=0; e = g->elist[e].nextt)
        {
            if(g->vsel[g->elist[e].f] > 0)
                g->vsel[g->elist[e].f]--;

            if(g->vsel[g->elist[e].f] > -1 && g->vsel[g->elist[e].f] <= 1)
            {
                g->vsel[g->elist[e].f] = -1;
                height[curq] = height[i]+1;
#ifdef DEBUG
                fprintf(stderr, "the %d vertex in the queue is high %d\n", curq, height[v]+1);
#endif
                queue[curq++] = g->elist[e].f;
            }
        }
    }

    /* two cases: either height of the tree is odd, or even */
    if(height[g->v-2] == height[g->v-1])
    {
#ifdef DEBUG
        fprintf(stderr, "d: vertices %llu and %llu\n are of heights %d and %d\n",
                queue[(g->v)-2], queue[(g->v)-1], height[g->v-2], height[g->v-1]);
#endif
        int a = queue[(g->v)-2];
        int b = queue[(g->v)-1];
        
        /* subdivide the edge */
        g->v++;
        g->fr = realloc(g->fr, (g->v+1)*sizeof(ednum));
        g->to = realloc(g->to, (g->v+1)*sizeof(ednum));
        g->vsel = realloc(g->vsel, (g->v+1)*sizeof(int));
        g->leaves = realloc(g->leaves, (g->v+1)*sizeof(venum));
        g->fr[g->v] = 0;
        g->to[g->v] = 0;
#ifdef DEBUG
        fprintf(stderr, "%d %d\n", a, b);
#endif
        venum other = replace_edge(g, a, b, g->v);
        add_edge(g, other, g->v);
        swap = g->v;
    } else {
        swap = queue[g->v-1];
    }
#ifdef DEBUG
    fprintf(stderr, "Swapping %llu %llu\n", swap, (unsigned long long) 1);
#endif
    swap_vertices(g, swap, 1);
    free(queue);
    free(height);
}

/* evaluates the graph based on the k vertices
 * given.
 * assumes there is a global graph *G
 */
void induce(int *array, int k)
{
    graph *ng;
    ng = malloc(sizeof(graph));
    int *tr;
    tr = calloc(G->v+1, sizeof(int));
    init_graph(ng, k);
    //debug(ng);
    for(int i=0; i<k; i++)
    {
#ifdef DEBUG
        fprintf(stderr, "translating G->%d as local %d\n", array[i], i+1);
#endif
        tr[array[i]] = i+1;
    }
    for(int i=0; i<k; i++)
    {
        for(ednum e=G->fr[array[i]]; e!=0; e=G->elist[e].nextf)
        {
#ifdef DEBUG
            fprintf(stderr, "checking edge from %llu to %llu\n", G->elist[e].f, G->elist[e].t);
            fprintf(stderr, "trans of f is %d and trans of t is %d \n", tr[G->elist[e].f], tr[G->elist[e].t]);
#endif

            if(tr[G->elist[e].f] && tr[G->elist[e].t])
                add_edge(ng, tr[G->elist[e].f], tr[G->elist[e].t]);
        }
    }

    //debug(ng);

    if(valid(ng))
    {
#ifdef DEBUG
        fprintf(stderr, "Centralizing.\n");
        debug(ng);
#endif
        centralize(ng);
        output_tree(ng);
    }

    eraseg(ng);
    free(tr);
}

/* assumes there are elements 0...max */
void ktuples(int *array, int pos, int element, int max)
{
    int k = Treesize;
    /*fprintf(stderr, "Now inspecting %d-tuple:\n", k);*/
    /*print_tuple(array, k);*/
    
    if(max - element < (k-1) - pos)
        return;
    
    array[pos] = element;
    if((k-1) == pos)
    {
#ifdef DEBUG
        fprintf(stderr, "the new tuple is to be evaluated\n");
        print_tuple(array, k);
#endif
        induce(array, k);
    }

    else
    {
       ktuples(array, pos+1, element+1, max);
    } 
    if(element < max)
    {   
        ktuples(array, pos, element+1, max);
    }
}


int main(int argc, char** argv)
{
    FILE* fin;
    fin = fopen(argv[1], "r");
    if(fin == NULL)
    {
        fprintf(stderr, "Source file could not be opened.\n");
        return -1;
    }
    char t[256], o[256], p[256];
    char dir[256];
    strcpy(dir, argv[3]); strcat(dir, "/");
    Treesize = atoi(argv[2]);
    
    int *array;
    array = calloc(Treesize, sizeof(int));
    ull count = 0;
    
    ull max = 0;
    if(strcmp(argv[4],"ALL") != 0)
    {
        max = atoi(argv[4]);
#ifdef DEBUG
        fprintf(stderr, "max is %llu\n", max);
#endif
    }

    while(!feof(fin))
    {
        if(max != 0 && count >= max)
            break;
        count++;
        if(count % 100000 == 0)
            fprintf(stderr, "%llu\n", count);
        sprintf(t, "%llu", count);
        strcpy(o, dir); strcpy(p,dir);
        strcat(o, t); strcat (p,t);
        strcat(o, ".normal.txt"); strcat(p, ".subdiv.txt");
        
        Normalsized = fopen(o, "w");
        Subdivided = fopen(p, "w");

        G = loadmckay(fin);
#ifdef DEBUG
        debug(G);
#endif
        ktuples(array, 0, 1, G->v);
        fclose(Normalsized);
        fclose(Subdivided);
        eraseg(G);
    }
    printf("%llu graphs processed.\n", count);
    fclose(fin);
    free(array);
    return 0;
}

